use super::*;
use std::{
    path::Path,
};
use tokio::{
    fs::{
	OpenOptions,
    },
    prelude::*,
    stream::StreamExt,
};

mod tasklist;
#[macro_use]
mod progress;

/// Decode a loli from path
pub async fn decode(from: impl AsRef<Path>, to: impl AsRef<Path>, progress: &mut progress::CommandSender) -> Result<loli::Loli, error::Error>
{
    prog_send!(link progress.println("Mapping child"));
    let base = loli::BasedLoli::map(from)?;
    let bounds = base.calculate_bounds()?; // If server is returning error code, this will fail.
    prog_send!(link progress.println(format!("Decoding {} -> {} bytes", base.as_ref().len(), base.decoded_size())));
    let mut decoded = bounds.create_child(to)?;
    prog_send!(progress.println("Decoding..."));
    let sz = base.decode(&mut decoded)?;
    prog_send!(link progress.println(format!("Decode complete ({} bytes)", sz)));

    Ok(decoded)
}

/// Download a loli async
pub async fn perform(url: impl AsRef<str>, path: impl AsRef<Path>, progress: &mut progress::CommandSender) -> Result<(), error::Error>
{
    let url = url.as_ref();
    let path = path.as_ref();


    let mut resp = reqwest::get(url).await?;
    let len = resp.content_length();

    //prog_send!(link progress.push_task(&task));
    if let Some(len) = len {
	prog_send!(progress.bump_max(len));
    } else {
	prog_send!(progress.bump_max(1));
    }
    
    let mut file = OpenOptions::new()
	.create(true)
	.truncate(true)
	.write(true)
	.read(true)
	.open(path).await?;
    
    prog_send!(progress.println(format!("Starting download of {:?} bytes to {:?}", len, path)));
    let mut bytes = resp.bytes_stream();
    while let Some(buffer) = bytes.next().await {
	let slice = buffer?;
	let slice = slice.as_ref();
	file.write(slice).await?;

	if let Some(_) = len {
	    prog_send!(progress.bump(slice.len() as u64));
	}
    }
    if len.is_none() {
	prog_send!(progress.bump(1));
    }
    prog_send!(progress.println(format!("done for {}", url)));
    //prog_send!(link progress.pop_task(task));

    Ok(())
}

pub async fn work(conf: config::Config) -> Result<(), Box<dyn std::error::Error>>
{
    let rating = conf.rating;
    let mut children = Vec::with_capacity(conf.output.len());

    let prog = progress::AsyncProgressCounter::new("Initialising...", 1);
    let mut prog_writer = prog.writer();

    let prog = prog.host();
    
    for path in conf.output.into_iter()
    {
	let url = url::parse(&rating);
	let mut prog = prog_writer.clone_with(format!("-> {:?}", path));
	children.push(tokio::task::spawn(async move {
	    prog.println(format!("Starting download ({})...", url)).await.expect("fatal");
	    
	    let task = format!("{:?}", path); //TODO: Real task name
	    prog_send!(link unwind prog.push_task(&task));

	    let temp = tempfile::TempFile::new();
	    match perform(&url, &temp, &mut prog).await {
		Err(e) => panic!("Failed downloading {} -> {:?}: {}", url, temp, e), //TODO: Make real error handler
		Ok(_) => {
		    let path = match path {
			config::OutputType::File(file) => file,
			config::OutputType::Directory(dir) => unimplemented!(), //TODO: implement get hash to file
		    };
		    let loli = match decode(&temp, &path, &mut prog).await {
			Ok(v) => v,
			Err(e) => panic!("Failed decoding: {}", e),
		    };
		},
	    }

	    prog_send!(link prog.pop_task(task)); //TODO: Make sure we don't return gracefully before this is called.
	}));
    }

    prog_send!(link prog_writer.println("Children working..."));
    for child in children.into_iter()
    {
	match child.await {
	    Ok(v) => (),
	    Err(err) => {
		println!("Child failed: {}", err);
	    },
	}
    }
    prog_send!(link prog_writer.set_title(""));
    prog_send!(try link prog_writer.kill());
    prog.await.expect("mpsc fatal");
    
    Ok(())
}
